/*
 *  This file is part of the SNPack library.
 *  Copyright (c) Francesco Sambo <sambofra@dei.unipd.it>
 *
 *  SNPack is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  SNPack is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See 
 *  the GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 */

#ifndef _BYTEARRAY_H_
#define _BYTEARRAY_H_

/**
 * \file  bytearray.h
 * \brief Definition of the ByteArray class.
 */

#include <stdint.h>
#include <cstddef>
#include <vector>
#include <fstream>
#include <cstring> 
#include <algorithm>

// debug
#include <cstdio>

using std::vector;
using std::ifstream;

// bit mask for fast selection of n rightmost bits of a byte
static uint_fast8_t R_BITS[9] = {
	0, 1, 3, 7, 15, 31, 63, 127, 255 };

/// Utility class for packing binary numbers in a byte array.
class ByteArray{
 
 public:
	/// Construct an empty ByteArray and initialize to zero
	/**
	    @param s   Size of the ByteArray in Bytes
	 */
	ByteArray( size_t s );
	
	/// Construct an empty ByteArray and fill from ifstream
	/**
	    @param s	Size of the ByteArray in Bytes
	    @param in	Input file stream to fill the ByteArray
	 */
	ByteArray( size_t s, ifstream& in );
	
	/// Destructor
	~ByteArray();
	
	/// Copy constructor
	ByteArray( const ByteArray& ba );
	
	/// Copy assignment operator
	ByteArray& operator= ( ByteArray ba );
	
	/// Write a number in the leftmost available bits
	/**
   	 @param num    Number to be written
    	 @param nBits  Number of leftmost available bits to use
 	*/ 
	template <typename T> void write( T num, uint_fast8_t nBits );
		
	/// Read the next nBits bits and convert to integer
	/**
 	 	@param nBits  Number of bits to read
	*/
	uint_fast32_t read( uint_fast8_t nBits );
		
	uint8_t * array;  ///< Internal representation of the array
	size_t size;      ///< Array size in bytes
	
	/// Clear the ByteArray
	void reset();
	
	/// Fancy print the ByteArray content
	void print()
	{
		uint_fast32_t i;
		for( i = 0; i < size; i++ )
			printf("%02X ",array[i]);
		printf("\n");
	}
 
 private:
	size_t bytePos;
	uint_fast8_t bitPos;
	
	static const uint_fast8_t B_SIZE = 8;
};


inline ByteArray::ByteArray( size_t s ) :
	size(s), bytePos(0), bitPos(0)
{
	array = new uint8_t[size](); // zero-initialized
}

inline ByteArray::ByteArray( size_t s, ifstream& in ) :
	size(s), bytePos(0), bitPos(0)
{
	array = new uint8_t[size];
	in.read( reinterpret_cast<char*>(array), s );
}

inline ByteArray::~ByteArray()
{
	delete[] array;
}

inline ByteArray::ByteArray( const ByteArray& ba ) :
	size(ba.size), bytePos(ba.bytePos), bitPos(ba.bitPos)
{
	array = new uint8_t[size]();
	memcpy( array, ba.array, size );
}

inline ByteArray& ByteArray::operator= ( ByteArray ba )
{
	size = ba.size; 
	bytePos = ba.bytePos; 
	bitPos = ba.bitPos;
	std::swap( array, ba.array );
	return *this; 
}

inline
void ByteArray::reset()
{
	uint_fast32_t i;
	for( i = 0; i < size; i++ )
		array[i] = 0;
	bitPos = 0;
	bytePos = 0;
}


template <typename T> inline
void ByteArray::write( T num, uint_fast8_t nBits )
{
	uint_fast8_t remByte = B_SIZE - bitPos;
	if( nBits < remByte ) 
	{
		// _num_ shorter than the remainder of the byte
		array[bytePos] |= (uint8_t)(num << (remByte - nBits));
		bitPos += nBits;
	}
	else if ( nBits > remByte )
	{
		T bmask = 255; // ...0 1111 1111
		
		// fill the remainder of the byte
		array[bytePos++] |= (uint8_t)((num >> (nBits - remByte)) & bmask);
		
		int_fast8_t remNum = nBits - (remByte + B_SIZE);
		
		// fill entire bytes
		while( remNum > 0 )
		{
			array[bytePos++] = (uint8_t)((num >> remNum) & bmask);
			remNum -= B_SIZE;
		}
		
		// partially fill the last byte (note: _remNum_ is negative)
		array[bytePos] = (uint8_t)(num << (-remNum));
		bitPos = B_SIZE + remNum;
	}
	else // nBits == remByte
	{
		array[bytePos++] |= (uint8_t)(num);
		bitPos = 0;
	}		
}


inline
uint_fast32_t ByteArray::read( uint_fast8_t nBits )
{
	uint_fast8_t remByte = B_SIZE - bitPos;
	uint_fast32_t num;
	
	if( nBits < remByte ) 
	{
		// return value shorter than the remainder of the byte
		bitPos += nBits;
		num = ((array[bytePos] >> (remByte - nBits)) & R_BITS[nBits]); 
	}
	else
	{
		num = 0;
				
		// get the remainder of the byte, with the appropriate shift
		num |= (((uint_fast32_t)(array[bytePos++] & R_BITS[remByte])) << (nBits - remByte));
		
		int_fast8_t remNum = nBits - (remByte + B_SIZE);
		
		// get entire bytes
		while( remNum > 0 )
		{
			num |= (((uint_fast32_t)array[bytePos++]) << remNum);
			remNum -= B_SIZE;
		}
		
		// get part of the last byte (note: _remNum_ is negative)
		num |= (((uint_fast32_t)array[bytePos]) >> (-remNum));
		bitPos = B_SIZE + remNum;
	}
	return num;
}

#endif
