/*
 *  This file is part of the SNPack library.
 *  Copyright (c) Francesco Sambo <sambofra@dei.unipd.it>
 *
 *  SNPack is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  SNPack is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See 
 *  the GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 */

#include "data.h"

Data::Data( ifstream& in, Info& infoRef, uint8_t chrInd, uint8_t fileType ) :
	info(infoRef), chr(chrInd), nSubj(info.nSubj), nSnps(info.chrs[chr].nSnps),
	chrCode(info.chrs[chr].code), nnzInd(info.chrs[chr].nnzInd),
	snpCode(nSnps,1), cost(nSnps,ceil(nSubj/4.0)), snpSize((nSubj+3)/4), 
	snpFreq(nSnps,3), nBitSubjDiff(nSnps,4)
	// note, bitCost and snpCode defaulted to Code 1
{
	try
	{
		// allocate space for data
		SNPs = new uint8_t[ (size_t)nSnps * snpSize ];
		//std::cout << "Allocation done\n";
		//std::cout.flush();
	}
	catch( bad_alloc& e )
	{
		string ext(e.what());
		ext += " chromosome ";
		ext += info.chrs[chr].code;
		
		throw BadAlloc(ext.c_str());
	}
	
	// load data
	switch( fileType )
	{
		case BED :
			loadFromBed( in );
			// compute costs for each code and best code
			bestCodes();			
			break;
			
		case PCK :
			loadFromPck( in );
	} 
}

Data::~Data()
{
	delete[] SNPs;
}

void Data::loadFromBed( ifstream& in )
{
	// The stream in should already be open
	size_t ptrSNPs = 0;	// pointer to SNPs, always incremented
	size_t toRead = (size_t)nSnps * snpSize;
	uint_fast32_t ctrSubj = 0; // subject counter
	uint_fast32_t ctrSnps = 0; // snp counter
	
	uint8_t b,c,d,shift,val,i;
	
	/* read entire chromosome into SNPs, then convert it in place */
	in.read(	reinterpret_cast<char*>(SNPs), toRead );
	
	/* begin conversion */
	ctrSubj = 0;
	vector<uint_fast32_t> prevSubj(4,0);
	vector<uint_fast32_t> maxDist(4,0);
	for( ptrSNPs = 0; ptrSNPs < toRead; ptrSNPs++) 
	{
		// c = in.get();
		c = SNPs[ ptrSNPs ];
		b = 0x0;
		/* four patients in a byte block */
		for ( shift = 0; shift < 8; shift+=2 )
		{
			d = c>>shift & 0x3; // select single patient
			/* map d <- plink coded */
			switch( d )
			{
				case 3:		// AA -> 0
					val = 3;
					break;
				case 2:		// aA -> 1
					b = b | (0x1<<shift);
					snpFreq(ctrSnps,2)++;
					val = 2;
					break;
				case 0:		// aa -> 2
					b = b | (0x2<<shift);
					snpFreq(ctrSnps,1)++;
					val = 1;
					break;
				case 1:		// miss -> 3 
					b = b | (0x3<<shift);
					snpFreq(ctrSnps,0)++;
					val = 0;
					break;
			}					
			
			/* accumulate maximum distances */
			if( ctrSubj - prevSubj[val] > maxDist[val] )
				maxDist[val] = ctrSubj - prevSubj[val];
			prevSubj[val] = ctrSubj;
			
			ctrSubj++;
			if ( ctrSubj == nSubj ) // end of subjects for the SNP
			{
				// compute the number of bits for the maximum distance
				for( i = 0; i < 4; i++ )
				{
					if(maxDist[i] > 0) 
						nBitSubjDiff( ctrSnps, i ) = ceil( log(maxDist[i]+1) / log(2) );
					else
						nBitSubjDiff( ctrSnps, i ) = 1;
					prevSubj[i] = 0;
					maxDist[i] = 0;
				}
				
				ctrSubj = 0;
				ctrSnps++;
				break;
			}
		}
		SNPs[ptrSNPs] = b;	// copy entire byte
	}
}

void Data::bestCodes()
{
	uint_fast32_t i, c2, c3, c4, c5;
	for( i = 0; i < nSnps; i++ )
	{
		// code costs
		c2 = ceil((nBitSubjDiff(i,0) * snpFreq(i,0) + 
			nBitSubjDiff(i,1) * snpFreq(i,1) + 3 * info.nBitSubj +
			nBitSubjDiff(i,2) * snpFreq(i,2)) / 8.0) + 2;
		
		c3 = ceil((nBitSubjDiff(i,0) * snpFreq(i,0) + 2 * info.nBitSubj +
			nBitSubjDiff(i,1) * snpFreq(i,1) + nSubj) / 8.0) + 1;
		
		c4 = ceil((nBitSubjDiff(i,1) * snpFreq(i,1) + 
			nBitSubjDiff(i,2) * snpFreq(i,2) + 3 * info.nBitSubj +
			nBitSubjDiff(i,3) * (nSubj-snpFreq(i,0)-snpFreq(i,1)-snpFreq(i,2))) / 8.0) + 2;
		
		c5 = ceil((nBitSubjDiff(i,1) * snpFreq(i,1) + 2 * info.nBitSubj +
			nBitSubjDiff(i,2) * snpFreq(i,2) + nSubj) / 8.0) + 1;	
		
		// record best code
		if( c2 < cost[i] )
		{
			cost[i] = c2;
			snpCode[i] = 2;
		}
		if( c3 < cost[i] )
		{
			cost[i] = c3;
			snpCode[i] = 3;
		}
		if( c4 < cost[i] )
		{
			cost[i] = c4;
			snpCode[i] = 4;
		}
		if( c5 < cost[i] )
		{
			cost[i] = c5;
			snpCode[i] = 5;
		}
	}
}

void Data::getSnpC1(uint_fast32_t snp, ByteArray& array)
{
	uint_fast32_t j;
	
	// C1 indicator, 0010 0000
	array.write( 32, 8 );
	
	// copy without further coding
	for( j = snp*(snpSize); j < (snp+1)*snpSize; j++ )
		array.write(SNPs[j], 8);
}

void Data::getSnpC2(uint_fast32_t snp, ByteArray& array)
{
	uint_fast32_t j;
	
	// C2 indicator, 0100 0000
	array.write( 2, 3 );
	
	// vectors storing the indices of miss, homo minor and etero
	vector<uint_fast32_t> misInd;
	vector<uint_fast32_t> homInd;
	vector<uint_fast32_t> hetInd;
	for( j = 0; j < nSubj; j++ )
		switch( this->get(snp,j) )
		{
			case HOM_FREQ :
				break;
			case HET :
				hetInd.push_back(j);
				break;
			case HOM_RARE :
				homInd.push_back(j);
				break;
			default : // MISS
				misInd.push_back(j);
		}
	
	// bits for het, homo minor and miss
	array.write( nBitSubjDiff(snp,2), 5 );
	array.write( nBitSubjDiff(snp,1), 8 );
	array.write( nBitSubjDiff(snp,0), 8 );
	
	// number of miss, homo minor and hetero
	array.write( misInd.size(), info.nBitSubj );
	array.write( homInd.size(), info.nBitSubj ); 
	array.write( hetInd.size(), info.nBitSubj );
	
	// add differences between indices one at a time to array
	if( misInd.size() > 0 )
	{
		array.write( misInd[0], nBitSubjDiff(snp,0) );
		for( j = 1; j < misInd.size(); j++ )
			array.write( misInd[j] - misInd[j-1], nBitSubjDiff(snp,0) );
	}
	if( homInd.size() > 0 )
	{
		array.write( homInd[0], nBitSubjDiff(snp,1) );
		for( j = 1; j < homInd.size(); j++ )
			array.write( homInd[j] - homInd[j-1], nBitSubjDiff(snp,1) );
	}
	if( hetInd.size() > 0 )
	{
		array.write( hetInd[0], nBitSubjDiff(snp,2) );
		for( j = 1; j < hetInd.size(); j++ )
			array.write( hetInd[j] - hetInd[j-1], nBitSubjDiff(snp,2) );
	}
}

void Data::getSnpC3(uint_fast32_t snp, ByteArray& array)
{
	uint_fast32_t j;
	
	// C3 indicator, 0110 0000
	array.write( 3, 3 );
	
	// vectors storing the indices of miss and homo minor,
	// hetero are directly encoded
	vector<uint_fast32_t> misInd;
	vector<uint_fast32_t> homInd;
	
	// bits for homo minor and miss
	array.write( nBitSubjDiff(snp,1), 5 );
	array.write( nBitSubjDiff(snp,0), 8 );
	
	array.write( snpFreq(snp, 0), info.nBitSubj ); // miss
	array.write( snpFreq(snp, 1), info.nBitSubj ); // homo
	
	// begin writing hetero, while counting miss and homo
	for( j = 0; j < nSubj; j++ )
		switch( this->get(snp,j) )
		{
			case HOM_FREQ :
				array.write( 0, 1 );
				break;
			case HET :
				array.write( 1, 1 );
				break;
			case HOM_RARE :
				array.write( 0, 1 );
				homInd.push_back(j);
				break;
			default : // MISS
				array.write( 0, 1 );
				misInd.push_back(j);
		}
	
	// add indices of miss and homo one at a time to array
	if( misInd.size() > 0 )
	{
		array.write( misInd[0], nBitSubjDiff(snp,0) );
		for( j = 1; j < misInd.size(); j++ )
			array.write( misInd[j] - misInd[j-1], nBitSubjDiff(snp,0) );
	}
	if( homInd.size() > 0 )
	{
		array.write( homInd[0], nBitSubjDiff(snp,1) );
		for( j = 1; j < homInd.size(); j++ )
			array.write( homInd[j] - homInd[j-1], nBitSubjDiff(snp,1) );
	}
}

void Data::getSnpC4(uint_fast32_t snp, ByteArray& array)
{
	uint_fast32_t j;
	
	// C4 indicator, 1000 0000
	array.write( 4, 3 );
	
	// vectors storing the indices of homo major, homo minor and hetero
	vector<uint_fast32_t> homMajInd;
	vector<uint_fast32_t> homMinInd;
	vector<uint_fast32_t> hetInd;
	for( j = 0; j < nSubj; j++ )
		switch( this->get(snp,j) )
		{
			case HOM_FREQ :
				homMajInd.push_back(j);
				break;
			case HET :
				hetInd.push_back(j);
				break;
			case HOM_RARE :
				homMinInd.push_back(j);
				break;
			// do nothing for MISS
		}
	
	// bits for homo major, hetero and homo minor
	array.write( nBitSubjDiff(snp,3), 5 );
	array.write( nBitSubjDiff(snp,2), 8 );
	array.write( nBitSubjDiff(snp,1), 8 );
	
	// number of homo major, homo minor and hetero
	array.write( homMajInd.size(), info.nBitSubj );
	array.write( homMinInd.size(), info.nBitSubj ); 
	array.write( hetInd.size(), info.nBitSubj );
	
	// add indices one at a time to array
	if( homMajInd.size() > 0 )
	{
		array.write( homMajInd[0], nBitSubjDiff(snp,3) );
		for( j = 1; j < homMajInd.size(); j++ )
			array.write( homMajInd[j] - homMajInd[j-1], nBitSubjDiff(snp,3) );
	}
	if( homMinInd.size() > 0 )
	{
		array.write( homMinInd[0], nBitSubjDiff(snp,1) );
		for( j = 1; j < homMinInd.size(); j++ )
			array.write( homMinInd[j] - homMinInd[j-1], nBitSubjDiff(snp,1) );
	}
	if( hetInd.size() > 0 )
	{
		array.write( hetInd[0], nBitSubjDiff(snp,2) );
		for( j = 1; j < hetInd.size(); j++ )
			array.write( hetInd[j] - hetInd[j-1], nBitSubjDiff(snp,2) );
	}
}

void Data::getSnpC5(uint_fast32_t snp, ByteArray& array)
{
	uint_fast32_t j;
	
	// C3 indicator, 1010 0000
	array.write( 5, 3 );
	
	// vectors storing the indices of homo minor and hetero,
	// miss are directly encoded
	vector<uint_fast32_t> homInd;
	vector<uint_fast32_t> hetInd;
	
	// bits for het and homo minor
	array.write( nBitSubjDiff(snp,2), 5 );
	array.write( nBitSubjDiff(snp,1), 8 );
	
	array.write( snpFreq(snp, 1), info.nBitSubj ); // homo
	array.write( snpFreq(snp, 2), info.nBitSubj ); // het
	
	// begin writing miss, while counting homo and hetero
	for( j = 0; j < nSubj; j++ )
		switch( this->get(snp,j) )
		{
			case HOM_FREQ :
				array.write( 0, 1 );
				break;
			case HET :
				array.write( 0, 1 );
				hetInd.push_back(j);
				break;
			case HOM_RARE :
				array.write( 0, 1 );
				homInd.push_back(j);
				break;
			default : // MISS
				array.write( 1, 1 );
		}
	
	// add indices of homo and het one at a time to array
	if( homInd.size() > 0 )
	{
		array.write( homInd[0], nBitSubjDiff(snp,1) );
		for( j = 1; j < homInd.size(); j++ )
			array.write( homInd[j] - homInd[j-1], nBitSubjDiff(snp,1) );
	}
	if( hetInd.size() > 0 )
	{
		array.write( hetInd[0], nBitSubjDiff(snp,2) );
		for( j = 1; j < hetInd.size(); j++ )
			array.write( hetInd[j] - hetInd[j-1], nBitSubjDiff(snp,2) );
	}
}

void Data::getSnpAs(uint_fast32_t snp, uint_fast32_t baseSnp, ByteArray& array)
{
	uint_fast32_t j;
	// "As" indicator, 000S XXXX XXXX XXXX XXXX XXXX
	// S : sign of the position of the base snp (0 right, 1 left)
	// X..X : index distance from the base
	array.write( 0, 3 );
	if( baseSnp > snp )
	{
		array.write( 0, 1 );
		array.write( baseSnp - snp, 20 );
	}
	else
	{
		array.write( 1, 1 );
		array.write( snp - baseSnp, 20 );
	}
	
	// collect differences w.r.t baseSnp
	vector<uint_fast32_t> diffInd;
	vector<uint_fast32_t> diffVal;

	saveDiff(snp, baseSnp, diffInd, diffVal);
	
	// write difference size and pairs (index,value)
	array.write( diffInd.size(), info.nBitSubj );
	for( j = 0; j < diffInd.size(); j++ )
	{
		array.write( diffInd[j], info.nBitSubj );
		array.write( diffVal[j], 2 );
	}
}

void Data::setSnpC1(uint_fast32_t snp, ByteArray& array)
{
	uint_fast32_t j;
		
	// copy without decoding, skipping C1 indicator
	for( j = 0; j < snpSize; j++ )
		SNPs[snp*snpSize + j] = array.array[j + 1];
}

void Data::setSnpC2(uint_fast32_t snp, ByteArray& array)
{
	// the C2 indicator is assumed to be already read, 
	// read the number of bits for each difference
	uint_fast32_t nBitHet = array.read( 5 );
	uint_fast32_t nBitHom = array.read( 8 );
	uint_fast32_t nBitMis = array.read( 8 );
	uint_fast32_t j;
	
	// set snp to 0, i.e. all homo frequent allele
	for( j = 0; j < snpSize; j++ )
		SNPs[snp*snpSize + j] = 0;
	
	// get the number of miss, homo rare and hetero
	uint_fast32_t nMis = array.read( info.nBitSubj );
	uint_fast32_t nHom = array.read( info.nBitSubj );
	uint_fast32_t nHet = array.read( info.nBitSubj );
	
	// read the indices and set the values
	uint_fast32_t misInd, homInd, hetInd;
	if( nMis > 0 )
	{
		misInd = array.read( nBitMis );
		this->set( snp, misInd, MISS);
		for( j = 1; j < nMis; j++ )
		{
			misInd += array.read( nBitMis );
			this->set( snp, misInd, MISS);
		}
	}
	if( nHom > 0 )
	{
		homInd = array.read( nBitHom );
		this->set( snp, homInd, HOM_RARE);
		for( j = 1; j < nHom; j++ )
		{
			homInd += array.read( nBitHom );
			this->set( snp, homInd, HOM_RARE);
		}
	}
	if( nHet > 0 )
	{
		hetInd = array.read( nBitHet );
		this->set( snp, hetInd, HET );
		for( j = 1; j < nHet; j++ )
		{
			hetInd += array.read( nBitHet );
			this->set( snp, hetInd, HET );
		}
	}
}

void Data::setSnpC3(uint_fast32_t snp, ByteArray& array)
{
	// the C3 indicator is assumed to be already read, 
	// read the number of bits for each difference
	uint_fast32_t nBitHom = array.read( 5 );
	uint_fast32_t nBitMis = array.read( 8 );
	uint_fast32_t j;
	
	// set snp to 0, i.e. all homo frequent allele
	for( j = 0; j < snpSize; j++ )
		SNPs[snp*snpSize + j] = 0;
	
	// get the number of miss and homo rare
	uint_fast32_t nMis = array.read( info.nBitSubj );
	uint_fast32_t nHom = array.read( info.nBitSubj );
	
	// read hetero bits and set the values
	for( j = 0; j < nSubj; j++ )
		if( array.read( 1 ) )
			this->set( snp, j, HET );
	
	// read the indices of mis and hom and set the values
	uint_fast32_t misInd, homInd;	
	if( nMis > 0 )
	{
		misInd = array.read( nBitMis );
		this->set( snp, misInd, MISS );
		for( j = 1; j < nMis; j++ )
		{
			misInd += array.read( nBitMis );
			this->set( snp, misInd, MISS );
		}
	}
	if( nHom > 0 )
	{
		homInd = array.read( nBitHom );
		this->set( snp, homInd, HOM_RARE );
		for( j = 1; j < nHom; j++ )
		{
			homInd += array.read( nBitHom );
			this->set( snp, homInd, HOM_RARE );
		}
	}
}

void Data::setSnpC4(uint_fast32_t snp, ByteArray& array)
{
	// the C4 indicator is assumed to be already read, 
	// read the number of bits for each difference
	uint_fast32_t nBitHomMaj = array.read( 5 );
	uint_fast32_t nBitHet = array.read( 8 );
	uint_fast32_t nBitHomMin = array.read( 8 );
	uint_fast32_t j;
	
	// set snp to 11, i.e. all miss
	for( j = 0; j < nSubj/4; j++ )
		SNPs[snp*snpSize + j] = 255;		// 1111 1111
	// last byte, if any
	switch( nSubj%4 )
	{
		case 1:
			SNPs[snp*snpSize + j] = 3;		// 0000 0011
			break;
		case 2:
			SNPs[snp*snpSize + j] = 15;	// 0000 1111
			break;
		case 3:
			SNPs[snp*snpSize + j] = 63;	// 0011 1111
	}
	
	// get the number of homo major, homo minor and hetero
	uint_fast32_t nHomMaj = array.read( info.nBitSubj );
	uint_fast32_t nHomMin = array.read( info.nBitSubj );
	uint_fast32_t nHet = array.read( info.nBitSubj );
	
	// read the indices and set the values
	uint_fast32_t homMajInd, homMinInd, hetInd;	
	if( nHomMaj > 0 )
	{
		homMajInd = array.read( nBitHomMaj );
		this->set( snp, homMajInd, HOM_FREQ);
		for( j = 1; j < nHomMaj; j++ )
		{
			homMajInd += array.read( nBitHomMaj );
			this->set( snp, homMajInd, HOM_FREQ );
		}
	}
	if( nHomMin > 0 )
	{
		homMinInd = array.read( nBitHomMin );
		this->set( snp, homMinInd, HOM_RARE );
		for( j = 1; j < nHomMin; j++ )
		{
			homMinInd += array.read( nBitHomMin );
			this->set( snp, homMinInd, HOM_RARE );
		}
	}
	if( nHet > 0 )
	{
		hetInd = array.read( nBitHet );
		this->set( snp, hetInd, HET );
		for( j = 1; j < nHet; j++ )
		{
			hetInd += array.read( nBitHet );
			this->set( snp, hetInd, HET );
		}
	}
}

void Data::setSnpC5(uint_fast32_t snp, ByteArray& array)
{
	// the C5 indicator is assumed to be already read, 
	// read the number of bits for each difference
	uint_fast32_t nBitHet = array.read( 5 );
	uint_fast32_t nBitHom = array.read( 8 );
	
	uint_fast32_t j;
	
	// set snp to 0, i.e. all homo frequent allele
	for( j = 0; j < snpSize; j++ )
		SNPs[snp*snpSize + j] = 0;
	
	// get the number of homo rare and hetero
	uint_fast32_t nHom = array.read( info.nBitSubj );
	uint_fast32_t nHet = array.read( info.nBitSubj );
	
	// read miss bits and set the values
	for( j = 0; j < nSubj; j++ )
		if( array.read( 1 ) )
			this->set( snp, j, MISS );
	
	// read the indices of hom and het and set the values
	uint_fast32_t hetInd, homInd;
	if( nHom > 0 )
	{
		homInd = array.read( nBitHom );
		this->set( snp, homInd, HOM_RARE);
		for( j = 1; j < nHom; j++ )
		{
			homInd += array.read( nBitHom );
			this->set( snp, homInd, HOM_RARE);
		}
	}
	if( nHet > 0 )
	{
		hetInd = array.read( nBitHet );
		this->set( snp, hetInd, HET);
		for( j = 1; j < nHet; j++ )
		{
			hetInd += array.read( nBitHet );
			this->set( snp, hetInd, HET);
		}
	}
}

void Data::setSnpAs(uint_fast32_t snp, ByteArray& array)
{
	// the As indicator is assumed to be already read
	
	uint_fast32_t j;
	uint_fast32_t baseSnp, nDiff, diffInd;
	
	// read sign and distance of the base SNP
	if( array.read(1) ) // sign bit
		baseSnp = snp - array.read(20);
	else
		baseSnp = snp + array.read(20);
	
	// copy from base SNP
	for( j = 0; j < snpSize; j++ )
		SNPs[snp*snpSize + j] = SNPs[baseSnp*snpSize + j];
	
	// change differences
	nDiff = array.read( info.nBitSubj );
	for( j = 0; j < nDiff; j++ )
	{
		diffInd = array.read( info.nBitSubj );
		(*this).set( snp, diffInd, array.read(2) );
	}
}

void Data::loadFromPck( ifstream& in )
{
	uint_fast32_t i,j;
	
	// retrieve chromosome size and the size of each packed SNP
	// std::cout << "Retrieve chromosome size\n";
	// std::cout.flush();
	uint_fast32_t chr_size = 0;
	for( j = 0; j < 4; j++ ) // pedantic, to get rid of endianness problems
		chr_size += (static_cast<uint_fast32_t>(in.get()) << ((3-j)*8));
	
	// std::cout << "Retrieve SNPs size\n";
	// std::cout.flush();
	vector<uint_fast32_t> sizes(nSnps,0); // all zeros
	for( i = 0; i < nSnps; i++ )
		for( j = 0; j < 4; j++ ) // pedantic, to get rid of endianness problems
			sizes[i] += (static_cast<uint_fast32_t>(in.get()) << ((3-j)*8));
		
	// First pass:
	// - read code bits for each SNP and decode SNPs with C1,2,3,4,5
	// - store summarized SNPs
	// std::cout << "Load C12345 SNPs\n";
	// std::cout.flush();
	
	vector<ByteArray*> baVec(nSnps);
	for( i = 0; i < nSnps; i++ )
	{
		baVec[i] = new ByteArray( sizes[i], in );
		snpCode[i] = baVec[i]->read(3);
		switch( snpCode[i] )
		{
			case 1 :
				(*this).setSnpC1( i, *(baVec[i]) );
				delete baVec[i];
				break;
			case 2 :
				(*this).setSnpC2( i, *(baVec[i]) );
				delete baVec[i];
				break;
			case 3 :
				(*this).setSnpC3( i, *(baVec[i]) );
				delete baVec[i];
				break;
			case 4 :
				(*this).setSnpC4( i, *(baVec[i]) );
				delete baVec[i];
				break;
			case 5 :
				(*this).setSnpC5( i, *(baVec[i]) );
				delete baVec[i];
				break;
		}
	}
	
	// for( i = 0; i < 33; i++)
	//	std::cout << (int)snpCode[i] << "\n";
		
	// Second pass:
	// - decode summarized SNPs
	// std::cout << "Load summarized SNPs\n";
	// std::cout.flush();
	for( i = 0; i < nSnps; i++ )
		if( snpCode[i] == 0 )
		{
			(*this).setSnpAs( i, *(baVec[i]) );
			delete baVec[i];
		}
		
}

void Data::saveToBed( ofstream& out )
{
	uint_fast32_t i,j,mult;
	uint8_t b,d;
	uint_fast8_t shift;
	
	// fill each SNP in the buffer and write it to disk
	char buffer[snpSize];
	for( i = 0; i < nSnps; i++)
	{
		// reset buffer
		for( j = 0; j < snpSize; j++ )
			buffer[j] = 0;
		
		// fill buffer
		mult = i*snpSize;
		for( j = 0; j < nSubj/4; j++ )
		{
			b = 0x0;
			for ( shift = 0; shift < 8; shift+=2 )
			{
				d = (SNPs[mult+j]>>shift) & 0x3; // select single patient
				switch( d )
				{
					case HOM_FREQ :
						b = b | (0x3<<shift);
						break;
					case HET :
						b = b | (0x2<<shift);
						break;
					case HOM_RARE :
						break;
					default : // MISS	
						b = b | (0x1<<shift);
				}
			}
			buffer[j] = b;
		}
		
		// possible last elements
		b = 0x0;
		for ( shift = 0; shift < 2*(nSubj%4); shift+=2 )
		{
			d = (SNPs[mult+j]>>shift) & 0x3; // select single patient
			switch( d )
			{
				case HOM_FREQ :
					b = b | (0x3<<shift);
					break;
				case HET :
					b = b | (0x2<<shift);
					break;
				case HOM_RARE :
					break;
				default : // MISS	
					b = b | (0x1<<shift);
			}
		}
		buffer[j] = b;
		
		// write to file
		out.write( buffer, snpSize );	
	}
}
