/*
 *  This file is part of the SNPack library.
 *  Copyright (c) Francesco Sambo <sambofra@dei.unipd.it>
 *
 *  SNPack is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  SNPack is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See 
 *  the GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 */

#ifndef __DATA_H__
#define __DATA_H__

/**
 * \file  data.h
 * \brief Definition of the Data class.
 */

#include <fstream>
#include <string>
#include <cstddef>
#include "info.h"
#include "matrix.h"
#include "bytearray.h"
#include <omp.h>

/* file types */
#define	BED	0
#define	PCK	1 

/* genotype codes */
#define HOM_FREQ	0
#define HET			1
#define HOM_RARE	2
#define MISS		3	

// Lookup table for fast comparisons between SNPs
static uint_fast32_t lookup[256] = {
	0, 1, 1, 1, 1, 2, 2, 2, 1, 2, 2, 2, 1, 2, 2, 2,
	1, 2, 2, 2, 2, 3, 3, 3, 2, 3, 3, 3, 2, 3, 3, 3,
	1, 2, 2, 2, 2, 3, 3, 3, 2, 3, 3, 3, 2, 3, 3, 3,
	1, 2, 2, 2, 2, 3, 3, 3, 2, 3, 3, 3, 2, 3, 3, 3,
	1, 2, 2, 2, 2, 3, 3, 3, 2, 3, 3, 3, 2, 3, 3, 3,
	2, 3, 3, 3, 3, 4, 4, 4, 3, 4, 4, 4, 3, 4, 4, 4,
	2, 3, 3, 3, 3, 4, 4, 4, 3, 4, 4, 4, 3, 4, 4, 4,
	2, 3, 3, 3, 3, 4, 4, 4, 3, 4, 4, 4, 3, 4, 4, 4,
	1, 2, 2, 2, 2, 3, 3, 3, 2, 3, 3, 3, 2, 3, 3, 3,
	2, 3, 3, 3, 3, 4, 4, 4, 3, 4, 4, 4, 3, 4, 4, 4,
	2, 3, 3, 3, 3, 4, 4, 4, 3, 4, 4, 4, 3, 4, 4, 4,
	2, 3, 3, 3, 3, 4, 4, 4, 3, 4, 4, 4, 3, 4, 4, 4,
	1, 2, 2, 2, 2, 3, 3, 3, 2, 3, 3, 3, 2, 3, 3, 3,
	2, 3, 3, 3, 3, 4, 4, 4, 3, 4, 4, 4, 3, 4, 4, 4,
	2, 3, 3, 3, 3, 4, 4, 4, 3, 4, 4, 4, 3, 4, 4, 4,
	2, 3, 3, 3, 3, 4, 4, 4, 3, 4, 4, 4, 3, 4, 4, 4};

static uint8_t zeroer[4] = {
	252,	// 1111 1100
	243,	// 1111 0011
	207,	// 1100 1111
	63};	// 0011 1111

using std::ifstream;
using std::ofstream;
using std::string;
using std::bad_alloc; 

/// Utilty class for reading from disk, storing and accessing SNP data 
/** This class implements all methods for loading data from disk in PLINK binary
    format (.bed, .bim, .fam) or SNPack format (.pck, .bim, .fam), access SNP
    data in RAM, code and decode SNPs with the 6 SNPack codes and write to disk
    in PLINK binary format.
	 
	 Internal representation: SNP major, 4 SNPs in one byte.
   
    SNP Coding: (improves speed!)
	 allele |   code
	 ------ | ---------
	   AA   | 0  ->  00
      aA   | 1  ->  01
      aa   | 2  ->  10
    miss   | 3  ->  11
	
    (note that plink coding is different!)
   
    Bit pairs are stored backwards, starting from the least significant pair of digits.
    The trailing byte for each row is (left) padded with zeros
**/ 
class Data
{
 friend class Pack;
 	
 public:
	/**
	 @param       in  Input file stream from which to read binary SNP data
	 @param  infoRef  Reference to an Info object, with information on the dataset
	 @param   chrInd  Index of the chromosome to load from disk
	 @param fileType  BED for binary PLINK, PCK for SNPack
	*/ 
	Data( ifstream& in, Info& infoRef, uint8_t chrInd, uint8_t fileType );
	~Data();

	Info& info;           ///< Info object, with information on the dataset
	uint8_t chr;          ///< Chromosome index
	uint32_t nSubj;       ///< Number of subjects
	uint32_t nSnps;       ///< Number of SNPs
	
	string chrCode;       ///< PLINK code of the chromosome
	uint_fast32_t nnzInd; ///< Index of the first SNP with nonzero pos
	
	vector<uint8_t> snpCode; ///< Best code for each SNP
	vector<uint32_t> cost;   ///< Cost of each SNP with the best code
	
	/// Save to output file stream in PLINK binary format
	void saveToBed( ofstream& out );
	
	/// Get a single genotype
	uint_fast8_t get(uint_fast32_t snp, uint_fast32_t subj);
	
	/// Set a single genotype
	void set(uint_fast32_t snp, uint_fast32_t subj, uint_fast8_t g);
	
	/// Number of differences between two snps
	uint_fast32_t diff(uint_fast32_t snp1, uint_fast32_t snp2);
	
	/// Given two SNPs, save indices and values of the differences
	void saveDiff(uint_fast32_t snp, uint_fast32_t baseSnp, 
		vector<uint_fast32_t>& diffInd, vector<uint_fast32_t>& diffVal);
	
	/// Number of Bytes of a SNP in SNPs
	uint_fast32_t snpSize;
	
	/// Set a SNP to all zeros
	void resetSNP( uint_fast32_t snp )
	{
		uint_fast32_t i;
		for( i = 0; i < snpSize; i++ )
			SNPs[snp*snpSize + i] = 0;
	}
	
	/// Print a fancy representation a SNP to standard output
	void printSNP( uint_fast32_t snp )
	{
		uint_fast32_t i;
		std::cout << "\n";
		for( i = 0; i < nSubj; i++ )
			std::cout << (unsigned int)this->get( snp, i ) << " ";
		std::cout << "\n";
	}

 private:
	
	/// Storage, SNP major 
	uint8_t * SNPs; 
	
	/// Compute the optimal cost and code for each snp
	void bestCodes();
	
	/// frequencies of miss, homo minor and hetero
	Matrix<uint_fast32_t> snpFreq;
	
	/// number of bits to represent differences between subject indices,
	/// for miss, homo minor, hetero and homo major
	Matrix<uint8_t> nBitSubjDiff;
	
	/* loaders */
	/// Load data in PLINK binary format from input file stream
	void loadFromBed( ifstream& in );
	/// Load data in SNPack format from input file stream
	void loadFromPck( ifstream& in );
	
  /** @name Getters of coded SNP data
   *  Functions to retrieve entire SNPs and save them to a ByteArray coded 
 	*  with the different codes.
   */
  ///@{
	/// Retrieve SNP, code with C1 and save to ByteArray	
	void getSnpC1(uint_fast32_t snp, ByteArray& array);
	/// Retrieve SNP, code with C2 and save to ByteArray	
	void getSnpC2(uint_fast32_t snp, ByteArray& array);
	/// Retrieve SNP, code with C3 and save to ByteArray	
	void getSnpC3(uint_fast32_t snp, ByteArray& array);
	/// Retrieve SNP, code with C4 and save to ByteArray	
	void getSnpC4(uint_fast32_t snp, ByteArray& array);
	/// Retrieve SNP, code with C5 and save to ByteArray	
	void getSnpC5(uint_fast32_t snp, ByteArray& array);

	/// Retrieve snp as summarized by baseSnp and save to ByteArray	
	void getSnpAs(uint_fast32_t snp, uint_fast32_t baseSnp, ByteArray& array);
	///@}

  /** @name Setters of coded SNP data
   *  Functions to save entire SNPs from a ByteArray, coded 
   *  with the different codes.
   */
  	///@{
	/// Save SNP contained in ByteArray, coded with C1
	void setSnpC1(uint_fast32_t snp, ByteArray& array);
	/// Save SNP contained in ByteArray, coded with C2
	void setSnpC2(uint_fast32_t snp, ByteArray& array);
	/// Save SNP contained in ByteArray, coded with C3
	void setSnpC3(uint_fast32_t snp, ByteArray& array);
	/// Save SNP contained in ByteArray, coded with C4
	void setSnpC4(uint_fast32_t snp, ByteArray& array);
	/// Save SNP contained in ByteArray, coded with C5
	void setSnpC5(uint_fast32_t snp, ByteArray& array);

	/// Save SNP contained in ByteArray, as summarized by a base SNP
	void setSnpAs(uint_fast32_t snp, ByteArray& array);
	///@}
};

/* Functions inlined for speed */

inline 
uint_fast8_t Data::get(uint_fast32_t snp, uint_fast32_t subj)
{
	#ifdef RANGE_CHECK
		if( snp >= nSnps || subj >= nSubj )
			throw BadIndex("Get: data subscript out of bounds");
	#endif
	uint8_t out = SNPs[snp*(snpSize) + subj/4]; // get the byte
	return (out>>(2*(subj%4))) & 0x3;	// shift and select 2 least significant bits
}

inline 
void Data::set(uint_fast32_t snp, uint_fast32_t subj, uint_fast8_t g)
{
	#ifdef RANGE_CHECK
		if( snp >= nSnps || subj >= nSubj )
			throw BadIndex("Set: data subscript out of bounds");
		if( g > 3 )
			throw BadValue("Set: invalid genotype code");
	#endif
	
	uint_fast32_t ind = snp*(snpSize) + subj/4;
	SNPs[ind] &= zeroer[subj%4];    // set target to zero
	SNPs[ind] |= (g << 2*(subj%4)); // assign the g value
}

inline
uint_fast32_t Data::diff(uint_fast32_t snp1, uint_fast32_t snp2)
{
	size_t i;
	uint_fast32_t df = 0;
	uint_fast8_t shift;
	
	size_t m1 = (size_t)snp1 * snpSize;
	size_t m2 = (size_t)snp2 * snpSize;
	
	// compare 4 patients at a time (1 Byte) with a XOR
	for(i = 0; i < nSubj/4; i++)
		df += lookup[ SNPs[m1+i] ^ SNPs[m2+i] ];
	
	// in case nSubj not a multiple of 4
	for( shift = 0; shift < nSubj%4; shift++ )
		if(((SNPs[m1+nSubj/4] ^ SNPs[m2+nSubj/4]) >> (2*shift)) & 0x3)
			df++;
	
	return df;
}

inline
void Data::saveDiff(uint_fast32_t snp, uint_fast32_t baseSnp, 
	vector<uint_fast32_t>& diffInd, vector<uint_fast32_t>& diffVal)
{
	size_t j;
	uint_fast8_t shift, tmp;
	
	size_t m1 = (size_t)snp * snpSize;
	size_t m2 = (size_t)baseSnp * snpSize;
	
	for( j = 0; j < nSubj/4; j++ )
	{
		for( shift = 0; shift < 4; shift++ )
		{
			tmp = ((SNPs[m1+j]>>(2*shift)) & 0x3);
			if( tmp != ((SNPs[m2+j]>>(2*shift)) & 0x3) )
			{
				diffInd.push_back(j*4+shift);
				diffVal.push_back(tmp);
			}
		}
	}
	// in case nSubj not a multiple of 4
	for( shift = 0; shift < nSubj%4; shift+=2 )
	{
		tmp = ((SNPs[m1+j]>>(2*shift)) & 0x3);
		if( tmp != ((SNPs[m2+j]>>(2*shift)) & 0x3) )
		{
			diffInd.push_back(j*4+shift);
			diffVal.push_back(tmp);
		}
	}
}


#endif