/*
 *  This file is part of the SNPack library.
 *  Copyright (c) Francesco Sambo <sambofra@dei.unipd.it>
 *
 *  SNPack is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  SNPack is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See 
 *  the GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 */

#include <iostream>
#include <cstdlib>
#include <fstream>
#include <vector>
#include <cstring>
#include <iomanip>
#include "info.h"
#include "data.h"
#include "util.h"
#include "bytearray.h"
#include "pack.h"

/**
 * \file  example.cpp
 * \brief Examples of use of the SNPack library.
 */

using std::cout;
using std::cin;

int main( int argc, char ** argv )
{
	try
	{
		uint_fast32_t chunk;
		
		cout << "Enter the number of chunks into which each chromosome ";
		cout << "should be split\n[0 for default: 4 * number_of_cores]:";
		cin >> chunk; 
		
		if( chunk == 0 )
		{
			chunk = 4;
				
			#ifdef _OPENMP
				chunk *= omp_get_max_threads(); 
			#endif
		}
		
		
		uint_fast32_t w;
		
		cout << "Enter the maximum neighbourhood extension in base pairs:";
		cin >> w;
		
		cout << "Loading data information\n";
		
		bool doNeigh = true;
		
		Info info("data/data.fam","data/data.bim",w,chunk,doNeigh);
			
		cout << "Subjects: " << info.nSubj << " SNPs: " << info.nSnps << "\n";
		uint_fast32_t i;
		cout << "Chromosomes: ";
		for( i = 0; i < info.nChrs; i++ )
			cout << info.chrs[i].code << " "; 
		cout << "\n\n"; 
		
		std::ifstream in("data/data.bed");
		/* drop first three bytes */
		for ( i = 0; i < 3; i++ )
			in.get();
		
		// cycle on chromosomes
		Data * pd;
		Pack * pp;
		std::ofstream out("tmp.pck");
		uint_fast32_t pckSize = 0, bedSize = 0;
		cout << "Compressing data, press any key to start\n";
		cin.get();

		for( i = 0; i < info.nChrs; i++ )
		{
			cout << "Loading chr " << info.chrs[i].code << "\n";
			cout.flush();
			pd = new Data(in, info, i, BED);
				
			cout << "Packing...\n";
			cout.flush();
			
			pp = new Pack( info, *pd );
			cout << "Saving to .pck\n";
			cout.flush();
			
			pckSize += pp->saveToPck( out );
			cout << "\n";
				
			bedSize += (pd->nSnps * ((pd->nSubj+3)/4));
				
			delete pp;
			delete pd;
		}
		
		cout << "Size .bed: " << std::setw(11) << bedSize << "\n";
		cout << "Size .pck: " << std::setw(11) << pckSize << "\n";
		cout << "Gain     : " << std::setw(11) << bedSize - pckSize << " (" << ((double)(bedSize - pckSize)/bedSize)*100 << "%)\n\n" ;
				
		in.close();
		out.close();
		
		cout << "Loading and converting data to .bed, press any key to start\n";
		cin.get();
		
		in.open("tmp.pck");
		out.open("tmp.bed");
		
		/* save first three bytes */
		// 01101100 00011011 00000001
		out.put(108);
		out.put(27);
		out.put(1);
			
		// cycle on chromosomes
		for( i = 0; i < info.nChrs; i++ )
		{
			cout << "Loading chr " << info.chrs[i].code << "\n";
			cout.flush();
			pd = new Data(in, info, i, PCK);
			
			cout << "\nFirst SNP of first subject in chromosome " << 
				info.chrs[i].code << ": ";
			switch( pd->get(1,1) ) // allele coding defined in data.h
			{
				case(HOM_FREQ) :
					cout << "homozygous for the major allele.\nPress any key to continue\n";
					break;
				case(HET) :
					cout << "heterozygous.\nPress any key to continue\n";
					break;
				case(HOM_RARE) :
					cout << "homozygous for the minor allele.\nPress any key to continue\n";
					break;
				default :
					cout << "missing value.\nPress any key to continue\n";		
			} 
			cin.get();
					
			cout << "Saving to .bed\n\n";
			cout.flush();
			pd->saveToBed( out );
				
			delete pd;
		}
				
		in.close();
		out.close();
		
		if( compare_files( "tmp.bed", "data/data.bed" ) )
		cout << "Data successfully converted to .bed file\n";
	}
	catch( BadIndex& e )
	{
		std::cerr << e.what() << std::endl;
		exit( EXIT_FAILURE );
	}
	catch( BadFileFormat& e )
	{
		std::cerr << e.what() << std::endl;
		exit( EXIT_FAILURE );
	}
	catch( BadAlloc& e )
	{
		std::cerr << e.what() << std::endl;
		exit( EXIT_FAILURE );
	}
	
	return 0;
}

