/*
 *  This file is part of the SNPack library.
 *  Copyright (c) Francesco Sambo <sambofra@dei.unipd.it>
 *
 *  SNPack is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  SNPack is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See 
 *  the GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 */

#include "info.h"

Chr::Chr( string c, vector<uint32_t>& pos, uint32_t maxGenDist, 
	uint_fast32_t nChunk, bool doNeigh ) :
	code(c), nSnps(pos.size()), nnzInd(0), szLeftNeigh(nSnps), 
	szRightNeigh(nSnps), ckStart(nChunk+1)
{
	if( doNeigh )
		computeNeigh( pos, maxGenDist, nChunk );
}

void Chr::computeNeigh( vector<uint32_t>& pos, uint32_t maxGenDist, uint_fast32_t nChunk )
{
	uint_fast32_t ind, ck, i;
	
	// skip SNPs with position 0
	while( pos[nnzInd] == 0 )
		nnzInd++;
	
	// subdivide what remains of the chromosome in chunks
	uint_fast32_t base = (nSnps-nnzInd) / nChunk;
	uint_fast32_t rem = (nSnps-nnzInd) % nChunk;
	vector<uint_fast32_t> sizes(nChunk, base);
	// redistribute remainder
	for(i = 0; i < rem; i++)
		sizes[i] += 1;
		
	// fill ckStart
	ckStart[0] = nnzInd;
	for( i = 0; i < nChunk; i++ )
		ckStart[i+1] = ckStart[i] + sizes[i];
	
	// compute neighboring intervals for each snp in each chunk
	#pragma omp parallel for schedule( dynamic ) private( ind )
	for( ck = 0; ck < nChunk; ck++ )
		for( ind = ckStart[ck]; ind < ckStart[ck+1]; ind++ )
		{
			// update left, double checking for begin of chunk
			if( ind - szLeftNeigh[ind] > ckStart[ck] )
				while( pos[ind] - pos[ind - szLeftNeigh[ind] - 1] <= maxGenDist )
				{
					szLeftNeigh[ind]++;
					if( ind - szLeftNeigh[ind] == ckStart[ck] )
						break;
				}
			
			// update right, double checking for end of chunk
			if( ind + szRightNeigh[ind] < ckStart[ck+1] - 1 )
				while( pos[ind + szRightNeigh[ind] + 1] - pos[ind] <= maxGenDist )
				{
					szRightNeigh[ind]++;
					if( ind + szRightNeigh[ind] == ckStart[ck+1] - 1 )
						break;
				}
		}
}

Info::Info( string famFileName, string bimFileName, uint32_t mgd, 
	uint_fast32_t nck, bool doNeigh ) :
	maxGenDist( mgd ), nChunk( nck )
{
	ifstream famFile;
	ifstream bimFile;
	string prevChr, nextChr, buf;
	vector<uint32_t> pos;
	uint32_t tmpPos;
		
	famFile.open(famFileName.c_str());
	if (!famFile)
	{
		cerr << "\nError opening " << famFileName << "\n";
		throw BadFileFormat("");
	}
	
	// count the number of subjects
	nSubj = 0;
	while ( getline(famFile, buf) )
	   ++nSubj;
	
	famFile.close();
	
	bimFile.open(bimFileName.c_str());
	if (!bimFile)
	{
		cerr << "\nError opening " << bimFileName << "\n";
		throw BadFileFormat("");
	}
	
	nBitSubj = (uint_fast8_t)ceil( log(nSubj) / log(2) );
	
	// build the _chrs_ structure
	bimFile >> prevChr;
	bimFile >> buf;			// SNP name, skip
	bimFile >> buf;			// cM position, skip
	bimFile >> tmpPos;
	bimFile >> buf >> buf;	// alleles, skip
	
	pos.push_back( tmpPos );
	
	nSnps = 1;
	nChrs = 0;
	bimFile >> nextChr;
	while(!bimFile.eof())
	{
		if( nextChr.compare(prevChr) != 0 )
		{
			// add a chromosome and reset pos
			chrs.push_back( Chr(prevChr, pos, maxGenDist, nChunk, doNeigh) );
			pos.clear();
			nChrs++;
		}
		
		bimFile >> buf;			// SNP name, skip
		bimFile >> buf;			// cM position, skip
		bimFile >> tmpPos;
		bimFile >> buf >> buf;	// alleles, skip
			
		pos.push_back( tmpPos );
		
		nSnps++;
		prevChr = nextChr;
		bimFile >> nextChr;
	}
	
	// store last chromosome
	chrs.push_back( Chr(prevChr, pos, maxGenDist, nChunk, doNeigh) );
	nChrs++;
	
	bimFile.close();
}

