/*
 *  This file is part of the SNPack library.
 *  Copyright (c) Francesco Sambo <sambofra@dei.unipd.it>
 *
 *  SNPack is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  SNPack is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See 
 *  the GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 */

#ifndef _INFO_H_
#define _INFO_H_

/**
 * \file  info.h
 * \brief Definition of the Info and Chr classes.
 */

#include <string>
#include <cstdio>
#include <stdint.h>
#include <vector>
#include <cmath>
#include <algorithm>
#include <fstream>
#include <iostream>
#include "util.h"

using std::vector;
using std::max;
using std::string;
using std::ifstream;
using std::getline;
using std::cerr;

/// Utility class for storing positional information on all SNPs of a chromosome
class Chr {
  public:
	/**
	 @param          c  PLINK numeric code of the chromosome
	 @param        pos  Vector of physical positions
	 @param maxGenDist  Maximum distance in bp between two neighbour SNPs
	 @param     nChunk  Number of chunks in which to split the chromosome
	 @param    doNeigh  Compute the neighbourhood of each SNP
	*/
	Chr( string c, vector<uint32_t>& pos, uint32_t maxGenDist, 
		uint_fast32_t nChunk, bool doNeigh );
	
	string code;          ///< PLINK numeric code of the chromosome
	uint_fast32_t nSnps;  ///< Number of SNPs in the chromosome
	uint_fast32_t nnzInd; ///< Index of the first SNP with nonzero position
	
	vector<uint32_t> szLeftNeigh;  ///< Size of the left neighbourhood for each SNP
	vector<uint32_t> szRightNeigh; ///< Size of the right neighbourhood for each SNP
	vector<uint32_t> ckStart;      ///< Starting indices of the chunks
	
  private:
	/// Compute the neighbourhood of each SNP
	void computeNeigh( vector<uint32_t>& pos, uint32_t maxGenDist, uint_fast32_t nChunk );
};

/// Utility class for loading and accessing information on the dataset
class Info {
  public:
	/**
	  @param  famFileName  Name of the .fam file
	  @param  bimFileName  Name of the .bim file
	  @param          mgd  Maximum distance in bp between two neighbour SNPs
	  @param          nck  Number of chunks in which to split each chromosome
	  @param      doNeigh  Compute neighbourhood of each SNP
	  */
	Info( string famFileName, string bimFileName, uint32_t mgd, 
		uint_fast32_t nck, bool doNeigh = true );
	
	uint32_t maxGenDist;   ///< Maximum distance in bp between two neighbour SNPs
	uint_fast32_t nChunk;  ///< Number of chunks in which to split each chromosome
	
	uint32_t nSubj;        ///< Number of subjects in the dataset
	uint32_t nSnps;        ///< Number of SNPs in the dataset
	uint8_t  nChrs;        ///< Number of chromosomes in the dataset
	
	uint_fast8_t nBitSubj; ///< Number of bits for representing the index of a subjects
	
	vector<Chr> chrs;      ///< Vector of chromosomes
};

#endif
