/*
 *  This file is part of the SNPack library.
 *  Copyright (c) Francesco Sambo <sambofra@dei.unipd.it>
 *
 *  SNPack is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  SNPack is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See 
 *  the GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 */

/**
 * \file  matrix.h
 * \brief Definition of the Matrix class and data structure.
 */

#ifndef _MATRIX_H_
#define _MATRIX_H_

#include <vector>
#include <iostream>

template<typename T>
class Matrix;

template <typename T> std::ostream& operator<<(std::ostream&, const Matrix<T>&);

/// Utility class for storing data in matrix format
template<typename T>
class Matrix {
 public:
	/// Initialize with number of rows and columns
   Matrix(unsigned long nrows, unsigned long ncols);

   /// Access method to get the (i,j) element:
   T&       operator() (unsigned long i, unsigned long j);
   /// Const access method to get the (i,j) element:
	const T& operator() (unsigned long i, unsigned long j) const;
   /// Access method to get the ith row:
   std::vector<T>&       operator() (unsigned long i);
   /// Const access method to get the ith row:
   const std::vector<T>& operator() (unsigned long i) const;

	/// Number of rows in this matrix
   unsigned long nrows() const;  
	/// Number of columns in this matrix
   unsigned long ncols() const;

   /// Fancy print method
   friend std::ostream& operator<< <T>(std::ostream&, const Matrix<T>&);

 private:
   std::vector<std::vector<T> > data;
};

template<typename T>
inline unsigned long Matrix<T>::nrows() const
{ return data.size(); }

template<typename T>
inline unsigned long Matrix<T>::ncols() const
{ return data[0].size(); }

template<typename T>
inline T& Matrix<T>::operator() (unsigned long row, unsigned long col)
{
  return data[row][col];
}

template<typename T>
inline const T& Matrix<T>::operator() (unsigned long row, unsigned long col) const
{
  return data[row][col];
}

template<typename T>
inline std::vector<T>& Matrix<T>::operator() (unsigned long row)
{
  return data[row];
}

template<typename T>
inline const std::vector<T>& Matrix<T>::operator() (unsigned long row) const
{
  return data[row];
}

template <typename T>
std::ostream& operator<<(std::ostream& o, const Matrix<T>& mat)
{
  o<<"\n";
  for(unsigned long i=0;i<mat.nrows();i++)
  {
    for(unsigned long j=0;j<mat.ncols();j++)
      o<<mat(i,j)<<"   ";
    o<<"\n\n";
  }
  o<<"\n";
  return o;
}

template<typename T>
Matrix<T>::Matrix(unsigned long nrows, unsigned long ncols)
   : data(nrows)
{
  for (unsigned long i = 0; i < nrows; ++i)
    data[i].resize(ncols);
}

#endif


