/*
 *  This file is part of the SNPack library.
 *  Copyright (c) Francesco Sambo <sambofra@dei.unipd.it>
 *
 *  SNPack is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  SNPack is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See 
 *  the GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 */

#include "pack.h"

Pack::Pack(Info& infoRef, Data& dataRef, bool verbose) :
	info(infoRef), data(dataRef), verb(verbose), nSubj(data.nSubj), nSnps(data.nSnps), 
	nChunk(info.nChunk), sLN(info.chrs[data.chr].szLeftNeigh), 
	sRN(info.chrs[data.chr].szRightNeigh), ckStart(info.chrs[data.chr].ckStart), 
	nBitSubj(info.nBitSubj), summary(nSnps,0), uNeigh(nSnps), uNeighDiff(nSnps), 
	gain(nSnps,0), byteCost(data.cost) 
{
	// no Packing for chr 0, only C12345
	if( data.chrCode.compare("0") != 0 )
	{
		#pragma omp parallel for schedule( dynamic )
		for( uint_fast32_t ck = 0; ck < nChunk; ck++ )
		{
			// std::cout << "Compute gain\n";
			computeGain( ck );
			
			// std::cout << "Compute summary\n";
			computeSummary( ck );
			
			if( verb )
				std::cout << "End chunk " << ck << "\n";			
		}
	}
}

void Pack::computeGain( uint_fast32_t ck )
{
	uint_fast32_t neighDiff, improved; 
	uint32_t  jit, oldjit; // for std::min to work
	int_fast32_t i, j, neighGain;
	vector<uint_fast32_t> neigh;
	
	// skip SNPs with pos = 0;
	for( i = ckStart[ck]; i < (int_fast32_t)ckStart[ck+1]; i++ )
	{
		// compute difference with an exponentially increasing number 
		// of snps, first on the left 
		jit = BASEJIT; oldjit = 0;
		improved = 1;
		while( oldjit < sLN[i] && improved )
		{
			improved = 0;
			for( j = i - oldjit - 1; j > i - (int_fast32_t)std::min(jit,sLN[i]) - 1; j-- )
			{
				neighDiff = data.diff(i, j);
				neighGain = data.cost[j] - ceil( (neighDiff*(nBitSubj + 2) + nBitSubj)/8.0 + 2 );
				if( neighGain > 0 )
				{
					improved = 1;
					uNeigh[i].push_back(j);
					uNeighDiff[i].push_back(neighDiff);
					gain[i] += neighGain;
				}		
			}
			// increase jitter
			oldjit = jit;
			jit *= 2;
		}
		// then on the right
		jit = BASEJIT; oldjit = 0;
		improved = 1;
		while( oldjit < sRN[i] && improved )
		{
			improved = 0;
			for( j = i + oldjit + 1; j <= i + (int_fast32_t)std::min(jit,sRN[i]); j++ )
			{
				neighDiff = data.diff(i, j);
				neighGain = data.cost[j] - ceil( (neighDiff*(nBitSubj + 2) + nBitSubj)/8.0 + 2 );
				if( neighGain > 0 )
				{
					improved = 1;
					uNeigh[i].push_back(j);
					uNeighDiff[i].push_back(neighDiff);
					gain[i] += neighGain;
				}		
			}
			// increase jitter
			oldjit = jit;
			jit *= 2;
		}
	}
}

void Pack::computeSummary( uint_fast32_t ck )
{
	uint_fast32_t i,j,k;
	uint_fast32_t m = max_element(gain.begin() + ckStart[ck], gain.begin() + ckStart[ck+1] - 1) - gain.begin();
	vector<uint32_t> uNeighM;
	vector<uint32_t>::iterator uNNMit;
	vector<bool> isBase(nSnps,0);
	
	// keep summarizing until effective
	while( gain[m] > 0 )
	{
		isBase[m] = true;
		gain[m] = 0;
		
		uNeighM.clear();
		
		// set m as base for uNeighs not yet in the summary (and not bases)
		for( i = 0; i < uNeigh[m].size(); i++ )
			if( (summary[uNeigh[m][i]] == 0) && (!isBase[uNeigh[m][i]]) )
			{
				gain[uNeigh[m][i]] = 0;
				summary[uNeigh[m][i]] = m - uNeigh[m][i];
				
				uNeighM.push_back(uNeigh[m][i]);
				
				// update the byte cost of the neighbour
				byteCost[uNeigh[m][i]] = ceil( (uNeighDiff[m][i]*(nBitSubj + 2) + nBitSubj)/8.0 + 2 );
			}
			
		/* update gain of all neighbors of _uNeighM_ and _m_ which
			1 - are not summarized neighbors of m, or m itself
			2 - have nonzero gain
			3 - have _uNeighM_ or _m_ as useful neighbor
		*/
		uNeighM.push_back(m);		
		for( i = 0; i < uNeighM.size(); i++)
		{
			for( j = uNeighM[i] - sLN[uNeighM[i]]; j <= uNeighM[i] + sRN[uNeighM[i]]; j++ )
				if( gain[j] > 0 && (find(uNeighM.begin(),uNeighM.end(),j) == uNeighM.end()) )
				{					
					uNNMit = find( uNeigh[j].begin(), uNeigh[j].end(), uNeighM[i] );
					if( uNNMit != uNeigh[j].end() )
					{
						// update gain
						k = uNNMit - uNeigh[j].begin();
						gain[j] -= (data.cost[uNeighM[i]] - ceil((uNeighDiff[j][k]*(nBitSubj + 2) + nBitSubj)/8.0 + 2));
					}
				}	
		}
		// next max
		m = max_element(gain.begin() + ckStart[ck], gain.begin() + ckStart[ck+1] - 1) - gain.begin();
	}
	// forcibly release some memory 
	for( i = ckStart[ck]; i < ckStart[ck+1]; i++ )
	{
		vector<uint32_t>().swap(uNeigh[i]);
		vector<uint32_t>().swap(uNeighDiff[i]);
	}
}

uint_fast32_t Pack::saveToPck( ofstream& out )
{
	uint_fast32_t i,j;
	ByteArray ba( data.snpSize + 1 ); // enough for storing all types of coding
	uint_fast32_t mask = 255; // 0..0 1111 1111
	
	// first save the byte size of the entire chromosome on disk,
	// for faster retrieval when loading
	uint_fast32_t chr_size = data.nSnps * 4; // the subsequent list of byte lengths
	for( i = 0; i < data.nSnps; i++ )
		chr_size += (byteCost[i] + 1);
	for( j = 0; j < 4; j++ ) // pedantic, to get rid of endianness problems
		out.put( (chr_size>>((3-j)*8)) & mask );
	
	// then save the byte length of each SNP, for faster retrieval when loading
	for( i = 0; i < data.nSnps; i++ )
		for( j = 0; j < 4; j++ ) // pedantic, to get rid of endianness problems
			out.put( ((byteCost[i] + 1)>>((3-j)*8)) & mask );
	
	// then save each SNP
	for( i = 0; i < data.nSnps; i++ )
	{
		ba.reset();
		if( summary[i] == 0 )
		{
			// base snp, best code among 1, 2, 3, 4 and 5
			switch( data.snpCode[i] )
			{
				case 1 :
					data.getSnpC1( i, ba );
					//std::cout << i << ": 1\n";
					break;
				case 2 :
					data.getSnpC2( i, ba );
					//std::cout << i << ": 2\n";
					break;
				case 3 :
					data.getSnpC3( i, ba );
					//std::cout << i << ": 3\n";
					break;
				case 4 :
					data.getSnpC4( i, ba );
					//std::cout << i << ": 4\n";
					break;
				default :
					data.getSnpC5( i, ba );
					//std::cout << i << ": 5\n";
			}
		}
		else // summarized SNP, code it wrt its base
		{
			data.getSnpAs( i, i + summary[i], ba );
		}
		out.write( reinterpret_cast<char*>(ba.array), byteCost[i] + 1 );		
	}
	
	return chr_size;
}

void Pack::writeStat( ofstream& stat )
{
	uint_fast32_t i;
	for( i = 0; i < data.nSnps; i++ )
	{
		stat << info.chrs[data.chr].code << "\t" << (int)(data.snpCode[i]) << "\t" <<
			summary[i] << "\t" <<  byteCost[i] + 1 << "\n";
	}
}
