/*
 *  This file is part of the SNPack library.
 *  Copyright (c) Francesco Sambo <sambofra@dei.unipd.it>
 *
 *  SNPack is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  SNPack is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See 
 *  the GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 */

#include "bytearray.h"
#include "info.h"
#include "data.h"
#include "pack.h"
#include "gtest/gtest.h"
#include <cstdio>
#include <algorithm>
#include <iterator>
#include <string>
#include <fstream>

/**
 * \file  test.cpp
 * \brief Definition and implementation of a unit test suite.
 */

#define BA_SIZE      2

namespace {

/// Fixture for testing class ByteArray.
class ByteArrayTest : public ::testing::Test {
 protected:
  
  ByteArrayTest() : emptyBA( BA_SIZE ) {
    // Automatically constructs emptyBA
  }

  ByteArray emptyBA;
};

TEST_F(ByteArrayTest, ConstructsByteArrayOfCorrectSize) {
	EXPECT_EQ(BA_SIZE,emptyBA.size);
}

TEST_F(ByteArrayTest, ConstructsEmptyByteArray) {
	for( int i = 0; i < BA_SIZE; i++ )
		EXPECT_EQ(0,emptyBA.array[i]);
}

TEST_F(ByteArrayTest, ConstructsFromIFStream) {
	std::ifstream ifs( "data/data.bed" );
	ByteArray ba( 3, ifs );
	EXPECT_EQ(108,ba.array[0]);
	EXPECT_EQ( 27,ba.array[1]);
	EXPECT_EQ(  1,ba.array[2]);
	ifs.close();
}

TEST_F(ByteArrayTest, WriteShortNum ) {
	ByteArray writeBA( 2 );
	//  53 = 00110101
	writeBA.write( 53, 6 );
	// 212 = 11010100 
	EXPECT_EQ(writeBA.array[0], 212);
	
	writeBA.write( 53, 6 );
	// 215 80 = 11010111 01010000
	EXPECT_EQ(writeBA.array[0], 215);
	EXPECT_EQ(writeBA.array[1], 80);
}

TEST_F(ByteArrayTest, WriteLongNum ) {
	ByteArray writeBA( 2 );
	// 1279 = 00000100 11111111
	writeBA.write( 1279, 12 );
	// 79 240 = 01001111 11110000 
	EXPECT_EQ(writeBA.array[0], 79);
	EXPECT_EQ(writeBA.array[1], 240);
}

TEST_F(ByteArrayTest, ReadShortNum ) {
	ByteArray readBA( 2 );
	// 212 = 11010100 
	readBA.array[0] = 212;
	//  53 = 110101
	EXPECT_EQ(readBA.read( 6 ), 53);
	readBA.reset();
	
	// 215 80 = 11010111 01010000
	readBA.array[0] = 215;
	readBA.array[1] =  80;
	EXPECT_EQ(readBA.read( 6 ), 53);
	EXPECT_EQ(readBA.read( 6 ), 53);
}

TEST_F(ByteArrayTest, ReadLongNum ) {
	ByteArray readBA( 2 );
	// 79 240 = 01001111 11110000 
	readBA.array[0] =  79;
	readBA.array[1] = 240;
	// 1279 = 00000100 11111111
	EXPECT_EQ(readBA.read( 12 ), 1279);
}

TEST_F(ByteArrayTest, CopyConstructor ) {
	ByteArray fromBA( 2 );
	// 1279 = 00000100 11111111
	fromBA.write( 1279, 12 );
	
	ByteArray toBA( fromBA );
	// 79 240 = 01001111 11110000 
	EXPECT_EQ(toBA.array[0], 79);
	EXPECT_EQ(toBA.array[1], 240);
	toBA.write( 10, 4 );
	// 79 250 = 01001111 11111010 
	EXPECT_EQ(toBA.array[1], 250);
}

TEST_F(ByteArrayTest, CopyAssignmentOperator ) {
	ByteArray fromBA( 2 );
	// 1279 = 00000100 11111111
	fromBA.write( 1279, 12 );
	
	ByteArray toBA( 2 );
	toBA.write( 216, 8 );
	toBA = fromBA;
	// 79 240 = 01001111 11110000 
	EXPECT_EQ(toBA.array[0], 79);
	EXPECT_EQ(toBA.array[1], 240);
	toBA.write( 10, 4 );
	// 79 250 = 01001111 11111010 
	EXPECT_EQ(toBA.array[1], 250);
}

/// Fixture for testing classes Info and Chr.
class InfoTest : public ::testing::Test {
 protected:
	 InfoTest() : pos(0)
	 {
		 int i;
		 // 10 inital SNPs with pos = 0
		 for( i = 0; i < 5; i++ )
			 pos.push_back( 0 );
		 // 5 SNPs spaced 10kb
		 for( i = 1e4; i <= 5e4; i+=1e4 )
			 pos.push_back( i );
		 // 5 SNPs spaced 20kb
		 for( i = 7e4; i <= 1.5e5; i+=2e4 )
			 pos.push_back( i );
	 }
	 
	 vector<uint32_t> pos;
};

// one chunk and no neighbourhood
TEST_F(InfoTest, ChrConstructor ) {
	Chr c( "1", pos, 0, 1, false);
	EXPECT_EQ( c.code, "1" );
	EXPECT_EQ( c.nSnps, 15 );
}	 

TEST_F(InfoTest, ChrConstrWithNeigh ) {
	Chr c( "1", pos, 5e4, 1, true);
	EXPECT_EQ( c.nnzInd, 5 );
	EXPECT_EQ( c.ckStart[0], 5 );
	int i;
	uint32_t expect_sln[15] = {0,0,0,0,0,0,1,2,3,4,4,3,2,2,2};
	uint32_t expect_srn[15] = {0,0,0,0,0,4,4,3,3,2,2,2,2,1,0}; 
	// first 5 SNPs, with pos = 0
	for ( i = 0; i < 15; i++ )
	{
		EXPECT_EQ( c.szLeftNeigh[i], expect_sln[i] );
		EXPECT_EQ( c.szRightNeigh[i], expect_srn[i] );
	}
}

TEST_F(InfoTest, ChrConstrWithNeighAndChunks ) {
	Chr c( "1", pos, 2e4, 3, true);
	EXPECT_EQ( c.ckStart[0], 5 );
	EXPECT_EQ( c.ckStart[1], 9 );
	EXPECT_EQ( c.ckStart[2], 12 );
	
	int i;
	uint32_t expect_sln[15] = {0,0,0,0,0,0,1,2,2,0,1,1,0,1,1};
	uint32_t expect_srn[15] = {0,0,0,0,0,2,2,1,0,1,1,0,1,1,0}; 
	// first 5 SNPs, with pos = 0
	for ( i = 0; i < 15; i++ )
	{
		EXPECT_EQ( c.szLeftNeigh[i], expect_sln[i] );
		EXPECT_EQ( c.szRightNeigh[i], expect_srn[i] );
	}
}

TEST_F(InfoTest, InfoConstructor ){
	Info in( "data/data.fam", "data/data.bim", 5e5, 4, true );
	EXPECT_EQ( in.maxGenDist, 5e5 );
	EXPECT_EQ( in.nChunk, 4 );
	EXPECT_EQ( in.nSubj, 2000 );
	EXPECT_EQ( in.nSnps, 3098 );
	EXPECT_EQ( in.nChrs, 22 );
	EXPECT_EQ( in.nBitSubj, 11 );
	EXPECT_EQ( in.chrs.size(), 22 );
	int i;
	uint32_t nspns[] = {140,139,143,140,139,142,140,142,140,141,
		143,140,138,142,139,142,142,138,143,140,142,143};
	char code[3];
	for( i = 0; i < in.nChrs; i++ )
	{
		EXPECT_EQ( in.chrs[i].nSnps, nspns[i] );
		sprintf(code,"%d",i+1);
		EXPECT_EQ( in.chrs[i].code, code );
	}
}

/// Fixture for testing class Data.
class DataTest : public ::testing::Test {
 protected:
	 DataTest() : info( "data/data.fam", "data/data.bim", 0, 1 ) 
	 {
	 }
	 
	 Info info;

};

TEST_F(DataTest, ConstructorBED){
	unsigned int i;
	
	ifstream in("data/data.bed");
	/* drop first three bytes */
	for ( i = 0; i < 3; i++ )
		in.get();
	
	Data d( in, info, 0, BED );
	EXPECT_EQ( d.chr, 0 );
	EXPECT_EQ( d.nSubj, 2000 );
	EXPECT_EQ( d.nSnps, 140 );
	EXPECT_EQ( d.chrCode, "1" );
	EXPECT_EQ( d.nnzInd, 0 );
	EXPECT_EQ( d.snpSize, 500 );
	EXPECT_EQ( d.snpCode[72], 1 );
	EXPECT_EQ( d.snpCode[73], 2 );
	EXPECT_EQ( d.snpCode[74], 3 );
	EXPECT_EQ( d.snpCode[75], 4 );
	EXPECT_EQ( d.snpCode[76], 5 );
	EXPECT_EQ( d.cost[72], 500 );
	EXPECT_EQ( d.cost[73], 454 );
	EXPECT_EQ( d.cost[74], 397 );
	EXPECT_EQ( d.cost[75], 332 );
	EXPECT_EQ( d.cost[76], 499 );
}

TEST_F(DataTest, ConstructorPCK){
	ifstream in("data/data.pck");
	
	Data d( in, info, 0, PCK );
	EXPECT_EQ( d.chr, 0 );
	EXPECT_EQ( d.nSubj, 2000 );
	EXPECT_EQ( d.nSnps, 140 );
	EXPECT_EQ( d.chrCode, "1" );
	EXPECT_EQ( d.nnzInd, 0 );
	EXPECT_EQ( d.snpSize, 500 );
}

TEST_F(DataTest, GetSet){
	unsigned int i;
	
	ifstream in("data/data.bed");
	/* drop first three bytes */
	for ( i = 0; i < 3; i++ )
		in.get();
	
	Data d( in, info, 0, BED );
	
	// get
	EXPECT_EQ( d.get(0,0), 0 );
	EXPECT_EQ( d.get(6,0), 1 );
	EXPECT_EQ( d.get(0,1), 2 );
	EXPECT_EQ( d.get(75,1999), 3 );
	
	// set
	d.set( 0, 0, 3 );
	d.set( 6, 0, 2 );
	d.set( 0, 1, 1 );
	d.set( 75, 1999, 0 );
	
	EXPECT_EQ( d.get(0,0), 3 );
	EXPECT_EQ( d.get(6,0), 2 );
	EXPECT_EQ( d.get(0,1), 1 );
	EXPECT_EQ( d.get(75,1999), 0 );
	
	// reset
	d.resetSNP( 0 );
	for( i = 0; i < d.nSubj; i++ )
		EXPECT_EQ( d.get(0,i), 0 );
}

TEST_F(DataTest, Diff){
	unsigned int i;
	
	ifstream in("data/data.bed");
	/* drop first three bytes */
	for ( i = 0; i < 3; i++ )
		in.get();
	
	Data d( in, info, 0, BED );
	
	// diff
	unsigned int nDiff = d.diff( 4, 5 );
	EXPECT_EQ( nDiff, 53 );
	
	// saveDiff
	vector< uint_fast32_t > diffInd;
	vector< uint_fast32_t > diffVal;
	d.saveDiff( 4, 5, diffInd, diffVal );
	EXPECT_EQ( diffInd[0], 258 );
	EXPECT_EQ( diffVal[0], 2 );
	EXPECT_EQ( diffInd[1], 265 );
	EXPECT_EQ( diffVal[1], 0 );
	EXPECT_EQ( diffInd[51], 1800 );
	EXPECT_EQ( diffVal[51], 0 );
	EXPECT_EQ( diffInd[52], 1801 );
	EXPECT_EQ( diffVal[52], 2 );
}

TEST_F(DataTest, SaveToBED){
	unsigned int i;
	
	ifstream in("data/data.bed");
	/* drop first three bytes */
	for ( i = 0; i < 3; i++ )
		in.get();
	
	ofstream out("tmp.bed");
	out.put(108);
	out.put(27);
	out.put(1);
	Data * d;
	for ( i = 0; i < info.nChrs; i++ )
	{
		d = new Data( in, info, i, BED );
		d -> saveToBed(out);
		delete d;
	}
	out.close();
	
	EXPECT_TRUE( compare_files("data/data.bed","tmp.bed") );
}

/// Fixture for testing class Pack.
class PackTest : public ::testing::Test {
 protected:
	 PackTest() : infoLin( "data/data.fam", "data/data.bim", 50000, 1 ),
	 	infoPar( "data/data.fam", "data/data.bim", 50000, 4 ),
	 	infoNoNeigh( "data/data.fam", "data/data.bim", 0, 1 )
	 {
	 }
	 
	 Info infoLin;
	 Info infoPar;
	 Info infoNoNeigh;
};

TEST_F(PackTest, SaveToPCKLinear){
	unsigned int i;
	
	ifstream in("data/data.bed");
	/* drop first three bytes */
	for ( i = 0; i < 3; i++ )
		in.get();
	
	// compress
	ofstream out("tmp.pck");
	Data * d;
	Pack * p;
	for ( i = 0; i < infoLin.nChrs; i++ )
	{
		d = new Data( in, infoLin, i, BED );
		p = new Pack( infoLin, *d, false );
		p -> saveToPck( out );
		delete d;
		delete p;
	}
	in.close();
	out.close();
	
	// decompress
	in.open("tmp.pck");
	out.open("tmp.bed");
	out.put(108);
	out.put(27);
	out.put(1);
	for ( i = 0; i < infoLin.nChrs; i++ )
	{
		d = new Data( in, infoLin, i, PCK );
		d -> saveToBed( out );
	}
	in.close();
	out.close();
	
	EXPECT_TRUE( compare_files("data/data.bed","tmp.bed") );
}
		
TEST_F(PackTest, SaveToPCKParallel){
	unsigned int i;
	
	ifstream in("data/data.bed");
	/* drop first three bytes */
	for ( i = 0; i < 3; i++ )
		in.get();
	
	// compress
	ofstream out("tmp.pck");
	Data * d;
	Pack * p;
	for ( i = 0; i < infoPar.nChrs; i++ )
	{
		d = new Data( in, infoPar, i, BED );
		p = new Pack( infoPar, *d, false );
		p -> saveToPck( out );
		delete d;
		delete p;
	}
	in.close();
	out.close();
	
	// decompress
	in.open("tmp.pck");
	out.open("tmp.bed");
	out.put(108);
	out.put(27);
	out.put(1);
	for ( i = 0; i < infoPar.nChrs; i++ )
	{
		d = new Data( in, infoPar, i, PCK );
		d -> saveToBed( out );
	}
	in.close();
	out.close();
	
	EXPECT_TRUE( compare_files("data/data.bed","tmp.bed") );
}	

TEST_F(PackTest, SaveToPCKNoNeigh){
	unsigned int i;
	
	ifstream in("data/data.bed");
	/* drop first three bytes */
	for ( i = 0; i < 3; i++ )
		in.get();
	
	// compress
	ofstream out("tmp.pck");
	Data * d;
	Pack * p;
	for ( i = 0; i < infoNoNeigh.nChrs; i++ )
	{
		d = new Data( in, infoNoNeigh, i, BED );
		p = new Pack( infoNoNeigh, *d, false );
		p -> saveToPck( out );
		delete d;
		delete p;
	}
	in.close();
	out.close();
	
	// decompress
	in.open("tmp.pck");
	out.open("tmp.bed");
	out.put(108);
	out.put(27);
	out.put(1);
	for ( i = 0; i < infoNoNeigh.nChrs; i++ )
	{
		d = new Data( in, infoNoNeigh, i, PCK );
		d -> saveToBed( out );
	}
	in.close();
	out.close();
	
	EXPECT_TRUE( compare_files("data/data.bed","tmp.bed") );
}

}  // namespace
